"""Defines items used in KiCad schematic files

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    19.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/
"""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional, List, Dict

from kiutils.items.common import Fill, Position, ColorRGBA, ProjectInstance, Stroke, Effects, Property
from kiutils.utils.strings import dequote

@dataclass
class Junction():
    """The ``junction`` token defines a junction in the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_junction_section
    """

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates of the junction"""

    diameter: float = 0
    """The ``diameter`` token attribute defines the DIAMETER of the junction. A diameter of 0
       is the default diameter in the system settings."""

    color: ColorRGBA = field(default_factory=lambda: ColorRGBA())
    """The ``color`` token attributes define the Red, Green, Blue, and Alpha transparency of
       the junction. If all four attributes are 0, the default junction color is used."""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Junction:
        """Convert the given S-Expresstion into a Junction object

        Args:
            - exp (list): Part of parsed S-Expression ``(junction ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not junction

        Returns:
            - Junction: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'junction':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'color': object.color = ColorRGBA().from_sexpr(item)
            if item[0] == 'diameter': object.color = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        uuid = f'\n{indents}  (uuid {self.uuid})\n' if self.uuid is not None else ''
        expression =  f'{indents}(junction (at {self.position.X} {self.position.Y}) (diameter {self.diameter}) {self.color.to_sexpr()}{uuid}{indents}){endline}'
        return expression

@dataclass
class NoConnect():
    """The ``no_connect`` token defines a unused pin connection in the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_no_connect_section
    """

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates of the no connect"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> NoConnect:
        """Convert the given S-Expresstion into a NoConnect object

        Args:
            - exp (list): Part of parsed S-Expression ``(no_connect ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not no_connect

        Returns:
            - NoConnect: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'no_connect':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        uuid = f' (uuid {self.uuid})' if self.uuid is not None else ''

        return f'{indents}(no_connect (at {self.position.X} {self.position.Y}){uuid}){endline}'

@dataclass
class BusEntry():
    """The ``bus_entry`` token defines a bus entry in the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_bus_entry_section
    """

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates of the bus entry"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    size: Position = field(default_factory=lambda: Position())         # Re-using Position class here
    """The ``size`` token attributes define the X and Y distance of the end point from
       the position of the bus entry"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the bus entry is drawn"""

    @classmethod
    def from_sexpr(cls, exp: list) -> BusEntry:
        """Convert the given S-Expresstion into a BusEntry object

        Args:
            - exp (list): Part of parsed S-Expression ``(bus_entry ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not bus_entry

        Returns:
            - BusEntry: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'bus_entry':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'size': object.size = Position().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(bus_entry (at {self.position.X} {self.position.Y}) (size {self.size.X} {self.size.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Connection():
    """The ``wire`` and ``bus`` tokens define wires and buses in the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_wire_and_bus_section
    """

    type: str = "wire"
    """The ``type`` token defines wether the connection is a ``bus`` or a ``wire``"""

    points: List[Position] = field(default_factory=list)
    """The ``points`` token defines the list of X and Y coordinates of start and end points
       of the wire or bus"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the connection is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Connection:
        """Convert the given S-Expresstion into a Connection object

        Args:
            - exp (list): Part of parsed S-Expression ``(wire ...)`` or ``(bus ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not wire or bus

        Returns:
            - Connection: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if not (exp[0] == 'wire' or exp[0] == 'bus'):
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.type = exp[0]
        for item in exp:
            if item[0] == 'pts':
                for point in item[1:]:
                    object.points.append(Position().from_sexpr(point))
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        points = ''
        for point in self.points:
            points += f' (xy {point.X} {point.Y})'

        expression =  f'{indents}({self.type} (pts{points})\n'
        expression += self.stroke.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class PolyLine():
    """The ``polyline`` token defines one or more lines that may or may not represent a polygon

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_graphical_line_section
    """

    points: List[Position] = field(default_factory=list)
    """The ``points`` token defines the list of X/Y coordinates of to draw line(s)
       between. A minimum of two points is required."""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the graphical line is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> PolyLine:
        """Convert the given S-Expresstion into a PolyLine object

        Args:
            - exp (list): Part of parsed S-Expression ``(polyline ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not polyline

        Returns:
            - PolyLine: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'polyline':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'pts':
                for point in item[1:]:
                    object.points.append(Position().from_sexpr(point))
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        points = ''
        for point in self.points:
            points += f' (xy {point.X} {point.Y})'

        expression =  f'{indents}(polyline (pts{points})\n'
        expression += self.stroke.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Text():
    """The ``text`` token defines graphical text in a schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_graphical_text_section
    """

    text: str = ""
    """The ``text`` token defines the text string"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the text"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Text:
        """Convert the given S-Expresstion into a Text object

        Args:
            - exp (list): Part of parsed S-Expression ``(text ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not text

        Returns:
            - Text: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'text':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''

        expression =  f'{indents}(text "{dequote(self.text)}"'

        # Strings longer or equal than 50 chars have the position in the next line
        if len(self.text) >= 50:
            expression += f'\n{indents}  '
        else:
            expression += ' '
        expression += f'(at {self.position.X} {self.position.Y}{posA})\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class TextBox():
    """The ``text_box`` token defines a text box inside a schematic

    Available since KiCad v7

    Documentation:
        ????
    """
    text: str = ""
    """The ``text`` token defines the text string"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the text"""

    size: Position = field(default_factory=lambda: Position())
    """The ``size`` token defines the size in X and Y direction. Angle is not used."""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` token defines the look of the outline of the text box"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token defines how the text box should be filled"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the text is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> TextBox:
        """Convert the given S-Expresstion into a TextBox object

        Args:
            - exp (list): Part of parsed S-Expression ``(text_box ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not text_box

        Returns:
            - TextBox: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'text_box':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'size': object.size = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''

        expression =  f'{indents}(text_box "{dequote(self.text)}"\n'
        expression += f'{indents}  (at {self.position.X} {self.position.Y}{posA}) (size {self.size.X} {self.size.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class LocalLabel():
    """The ``label`` token defines an wire or bus label name in a schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#local_label_section
    """

    text: str = ""
    """The ``text`` token defines the text in the label"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the label"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the label is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
    with the global label have been place automatically"""

    @classmethod
    def from_sexpr(cls, exp: list) -> LocalLabel:
        """Convert the given S-Expresstion into a LocalLabel object

        Args:
            - exp (list): Part of parsed S-Expression ``(label ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not label

        Returns:
            - LocalLabel: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'label':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        fieldsAutoplaced = ' (fields_autoplaced)' if self.fieldsAutoplaced else ''

        expression =  f'{indents}(label "{dequote(self.text)}" (at {self.position.X} {self.position.Y}{posA}){fieldsAutoplaced}\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class GlobalLabel():
    """The ``global_label`` token defines a label name that is visible across all schematics in a design

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_global_label_section
    """

    text: str = ""
    """The ``text`` token defines the text in the label"""

    shape: str = "input"
    """The ``shape`` token defines the way the global label is drawn. Possible values are:
       ``input``, ``output``, ``bidirectional``, ``tri_state``, ``passive``."""

    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
       with the global label have been place automatically"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the label"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the label is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    properties: List[Property] = field(default_factory=list)
    """	The ``properties`` token defines a list of properties of the global label. Currently, the
    only supported property is the inter-sheet reference"""

    @classmethod
    def from_sexpr(cls, exp: list) -> GlobalLabel:
        """Convert the given S-Expresstion into a GlobalLabel object

        Args:
            - exp (list): Part of parsed S-Expression ``(global_label ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not global_label

        Returns:
            - GlobalLabel: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'global_label':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'property': object.properties.append(Property().from_sexpr(item))
            if item[0] == 'shape': object.shape = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        fa = ' (fields_autoplaced)' if self.fieldsAutoplaced else ''

        expression =  f'{indents}(global_label "{dequote(self.text)}" (shape {self.shape}) (at {self.position.X} {self.position.Y}{posA}){fa}\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        for property in self.properties:
            expression += property.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class HierarchicalLabel():
    """The ``hierarchical_label`` token defines a label that are used by hierarchical sheets to
    define connections between sheet in hierarchical designs

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_label_section
    """

    text: str = ""
    """The ``text`` token defines the text in the label"""

    shape: str = "input"
    """The ``shape`` token defines the way the global label is drawn. Possible values are:
    ``input``, ``output``, ``bidirectional``, ``tri_state``, ``passive``."""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` token defines the X and Y coordinates and rotation angle of the label"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` token defines how the label is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""
    
    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
    with the global label have been place automatically"""

    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalLabel:
        """Convert the given S-Expresstion into a HierarchicalLabel object

        Args:
            - exp (list): Part of parsed S-Expression ``(hierarchical_label ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not hierarchical_label

        Returns:
            - HierarchicalLabel: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'hierarchical_label':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'shape': object.shape = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        fieldsAutoplaced = ' (fields_autoplaced)' if self.fieldsAutoplaced else ''

        expression =  f'{indents}(hierarchical_label "{dequote(self.text)}" (shape {self.shape}) (at {self.position.X} {self.position.Y}{posA}){fieldsAutoplaced}\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SymbolProjectPath():
    """The symbol project path defines the ``path`` token to the sheet instance of the instance data
    of a symbol.

    Available since KiCad v7.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_symbol_section
    """
    
    sheetInstancePath: str = ""
    """The ``PATH_INSTANCE`` token defines the path to the symbol instance"""

    reference: str = ""
    """The ``reference`` token is a string that defines the reference designator for the symbol
    instance"""

    unit: int = 1
    """The ``unit`` token is a integer that defines the symbol unit for the symbol instance. For 
    symbols that do not define multiple units, this will always be 1."""

    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolProjectPath:        
        """Convert the given S-Expression into a SymbolProjectPath object

        Args:
            - exp (list): Part of parsed S-Expression ``(path ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not path

        Returns:
            - SymbolProjectPath: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'path':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.sheetInstancePath = exp[1]
        for item in exp[2:]:
            if item[0] == 'reference': object.reference = item[1]
            if item[0] == 'unit': object.unit = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        expression =  f'{indents}(path "{dequote(self.sheetInstancePath)}"\n'
        expression += f'{indents}  (reference "{dequote(self.reference)}") (unit {self.unit})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SymbolProjectInstance(ProjectInstance):
    """The ``project`` token attribute defines the name of the project as well as a list of symbol
    project paths (instance data). There can be instance data from other project when schematics 
    are shared across multiple projects. The projects will have to be sorted by the ``name`` token
    in alphabetical order.

    Available since KiCad v7.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_symbol_section
    """

    paths: List[SymbolProjectPath] = field(default_factory=list)
    """The ``paths`` token defines a list of symbol project paths for this project instance"""
    
    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolProjectInstance:        
        """Convert the given S-Expression into a SymbolProjectInstance object

        Args:
            - exp (list): Part of parsed S-Expression ``(project ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not project

        Returns:
            - SymbolProjectInstance: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'project':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.name = exp[1]
        for item in exp[2:]:
            if item[0] == 'path': object.paths.append(SymbolProjectPath.from_sexpr(item))
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        expression = f'{indents}(project "{dequote(self.name)}"\n'
        for path in self.paths:
            expression += path.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class SchematicSymbol():
    """The ``symbol`` token in the symbol section of the schematic defines an instance of a symbol
    from the library symbol section of the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_symbol_section
    """

    @property
    def libId(self) -> str:
        """The ``lib_id`` token defines which symbol in the library symbol section of the schematic
        this schematic symbol references. In ``kiutils``, the ``lib_id`` token is a combination of
        both the ``libraryNickname`` and ``entryName`` token. Setting the ``lib_id`` token will
        update those tokens accordingly.

        Returns:
            - Symbol id in the following format: ``<libraryNickname>:<entryName>`` or ``<entryName>``,
              if ``libraryNickname`` token is not set.
        """
        if self.libraryNickname:
            return f'{self.libraryNickname}:{self.entryName}'
        else:
            return f'{self.entryName}'

    @libId.setter
    def libId(self, symbol_id: str):
        """Sets the ``lib_id`` token and parses its contents into the ``libraryNickname`` and
        ``entryName`` token.

        Args:
            - symbol_id (str): The symbol id in the following format: ``<libraryNickname>:<entryName>``
              or only ``<entryName>``
        """
        parse_symbol_id = re.match(r"^(.+?):(.+?)$", symbol_id)
        if parse_symbol_id:
            self.libraryNickname = parse_symbol_id.group(1)
            self.entryName = parse_symbol_id.group(2)
        else:
            self.libraryNickname = None
            self.entryName = symbol_id

    libraryNickname: Optional[str] = None
    """The optional ``libraryNickname`` token defines which symbol library this symbol belongs to
    and is a part of the ``id`` token"""

    entryName: str = None
    """The ``entryName`` token defines the actual name of the symbol and is a part of the ``id``
    token"""

    libName: Optional[str] = None
    """The optional ``lib_name`` token is only set when the symbol was edited in the schematic.
    It may be set to ``<entryName>_X`` where X is a unique number that specifies which variation
    this symbol is of its original."""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and angle of rotation of the symbol"""

    unit: Optional[int] = None
    """The optional ``unit`` token attribute defines which unit in the symbol library definition
    that the schematic symbol represents"""

    inBom: bool = False
    """The ``in_bom`` token attribute determines whether the schematic symbol appears in any bill
    of materials output"""

    onBoard: bool = False
    """The ``on_board`` token attribute determines if the footprint associated with the symbol is
    exported to the board via the netlist"""

    dnp: Optional[bool] = None
    """The optional ``dnp`` token defines if a symbol is marked as do-not-populate in the schematic. 
    
    Available since KiCad v7"""

    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
    with the global label have been place automatically"""

    uuid: Optional[str] = ""
    """The optional `uuid` defines the universally unique identifier"""

    properties: List[Property] = field(default_factory=list)
    """The ``properties`` section defines a list of symbol properties of the schematic symbol"""

    pins: Dict[str, str] = field(default_factory=dict)
    """The ``pins`` token defines a dictionary with pin numbers in form of strings as keys and
    uuid's as values"""

    mirror: Optional[str] = None
    """The ``mirror`` token defines if the symbol is mirrored in the schematic. Accepted values:
    ``x`` or ``y``. When mirroring around the x and y axis at the same time use some additional
    rotation to get the correct orientation of the symbol."""

    instances: List[SymbolProjectInstance] = field(default_factory=list)
    """The ``instances`` token defines a list of symbol instances grouped by project. Every symbol 
    will have a least one instance.
    
    Available since KiCad v7."""

    @classmethod
    def from_sexpr(cls, exp: list) -> SchematicSymbol:
        """Convert the given S-Expresstion into a SchematicSymbol object

        Args:
            - exp (list): Part of parsed S-Expression ``(symbol ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not symbol

        Returns:
            - SchematicSymbol: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'symbol':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
            if item[0] == 'lib_id': object.libId = item[1]
            if item[0] == 'lib_name': object.libName = item[1]
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'unit': object.unit = item[1]
            if item[0] == 'in_bom': object.inBom = True if item[1] == 'yes' else False
            if item[0] == 'on_board': object.onBoard = True if item[1] == 'yes' else False
            if item[0] == 'dnp': object.dnp = True if item[1] == 'yes' else False
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'property': object.properties.append(Property().from_sexpr(item))
            if item[0] == 'pin': object.pins.update({item[1]: item[2][1]})
            if item[0] == 'mirror': object.mirror = item[1]
            if item[0] == 'instances':
                for instance in item[1:]:
                    object.instances.append(SymbolProjectInstance.from_sexpr(instance))
        
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        fa = f' (fields_autoplaced)' if self.fieldsAutoplaced else ''
        inBom = 'yes' if self.inBom else 'no'
        onBoard = 'yes' if self.onBoard else 'no'
        mirror = f' (mirror {self.mirror})' if self.mirror is not None else ''
        unit = f' (unit {self.unit})' if self.unit is not None else ''
        lib_name = f' (lib_name "{dequote(self.libName)}")' if self.libName is not None else ''
        if self.dnp is not None:
            dnp = ' (dnp yes)' if self.dnp else ' (dnp no)'
        else:
            dnp = ''

        expression =  f'{indents}(symbol{lib_name} (lib_id "{dequote(self.libId)}") (at {self.position.X} {self.position.Y}{posA}){mirror}{unit}\n'
        expression += f'{indents}  (in_bom {inBom}) (on_board {onBoard}){dnp}{fa}\n'
        if self.uuid:
            expression += f'{indents}  (uuid {self.uuid})\n'
        for property in self.properties:
            expression += property.to_sexpr(indent+2)
        for number, uuid in self.pins.items():
            expression += f'{indents}  (pin "{dequote(number)}" (uuid {uuid}))\n'
        if len(self.instances) != 0:
            expression += f'{indents}  (instances\n'
            for instance in self.instances:
                expression += instance.to_sexpr(indent+4)
            expression += f'{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class HierarchicalPin():
    """The ``pin`` token in a sheet object defines an electrical connection between the sheet in a
       schematic with the hierarchical label defined in the associated schematic file

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_sheet_pin_definition
    """

    name: str = ""
    """	The ``name`` attribute defines the name of the sheet pin. It must have an identically named
        hierarchical label in the associated schematic file."""

    connectionType: str = "input"
    """The electrical connect type token defines the type of electrical connect made by the
       sheet pin"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and angle of rotation of the pin"""

    effects: Effects = field(default_factory=lambda: Effects())
    """The ``effects`` section defines how the pin name text is drawn"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalPin:
        """Convert the given S-Expresstion into a HierarchicalPin object

        Args:
            - exp (list): Part of parsed S-Expression ``(pin ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not pin

        Returns:
            - HierarchicalPin: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'pin':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.name = exp[1]
        object.connectionType = exp[2]
        for item in exp[3:]:
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'effects': object.effects = Effects().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''

        expression =  f'{indents}(pin "{dequote(self.name)}" {self.connectionType} (at {self.position.X} {self.position.Y}{posA})\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression


@dataclass
class HierarchicalSheetProjectPath():
    """The symbol project path defines the ``path`` token to the sheet instance of the instance data
    of a symbol.

    Available since KiCad v7.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_sheet_section
    """
    
    sheetInstancePath: str = ""
    """The ``PATH_INSTANCE`` token defines the path to the symbol instance"""

    page: str = ""
    """The ``page`` token is a string that defines the page number of the sheet instance"""

    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalSheetProjectPath:        
        """Convert the given S-Expression into a HierarchicalSheetProjectPath object

        Args:
            - exp (list): Part of parsed S-Expression ``(path ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not path

        Returns:
            - HierarchicalSheetProjectPath: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'path':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.sheetInstancePath = exp[1]
        for item in exp[2:]:
            if item[0] == 'page': object.page = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        return f'{indents}(path "{dequote(self.sheetInstancePath)}" (page "{dequote(self.page)}")){endline}'

@dataclass
class HierarchicalSheetProjectInstance(ProjectInstance):
    """The ``project`` token attribute defines the name of the project as well as a list of 
    hierarchical sheet project paths (instance data). There can be instance data from other project 
    when schematics are shared across multiple projects. The projects will have to be sorted by the 
    ``name`` token in alphabetical order.

    Available since KiCad v7.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_sheet_section
    """

    paths: List[HierarchicalSheetProjectPath] = field(default_factory=list)
    """The ``paths`` token defines a list of hierarchical sheet project paths for this project instance"""
    
    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalSheetProjectInstance:        
        """Convert the given S-Expression into a HierarchicalSheetProjectInstance object

        Args:
            - exp (list): Part of parsed S-Expression ``(project ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not project

        Returns:
            - HierarchicalSheetProjectInstance: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) < 2:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'project':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.name = exp[1]
        for item in exp[2:]:
            if item[0] == 'path': object.paths.append(HierarchicalSheetProjectPath.from_sexpr(item))
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        expression = f'{indents}(project "{dequote(self.name)}"\n'
        for path in self.paths:
            expression += path.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression

@dataclass
class HierarchicalSheet():
    """The ``sheet`` token defines a hierarchical sheet of the schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_sheet_section
    """

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and angle of rotation of the sheet in the schematic"""

    width: float = 0
    """The ``width`` token defines the width of the sheet"""

    height: float = 0
    """The ``height`` token defines the height of the sheet"""

    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
       with the global label have been place automatically"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the sheet outline is drawn"""

    fill: ColorRGBA = field(default_factory=lambda: ColorRGBA())
    """The fill defines the color how the sheet is filled"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    sheetName: Property = field(default_factory=lambda: Property(key="Sheet name"))
    """The ``sheetName`` is a property that defines the name of the sheet. The property's
       key should therefore be set to `Sheet name`"""

    fileName: Property = field(default_factory=lambda: Property(key="Sheet file"))
    """The ``fileName`` is a property that defines the file name of the sheet. The property's
       key should therefore be set to `Sheet file`"""

    pins: List[HierarchicalPin] = field(default_factory=list)
    """The ``pins`` section is a list of hierarchical pins that map a hierarchical label defined in
       the associated schematic file"""
    
    instances: List[HierarchicalSheetProjectInstance] = field(default_factory=list)
    """The ``instances`` token defines a list of hierachical sheet instances grouped by project. 
    Every hierarchical sheet will have a least one instance.
    
    Available since KiCad v7."""

    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalSheet:
        """Convert the given S-Expresstion into a HierarchicalSheet object

        Args:
            - exp (list): Part of parsed S-Expression ``(sheet ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not sheet

        Returns:
            - HierarchicalSheet: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'sheet':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp[1:]:
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'size':
                object.width = item[1]
                object.height = item[2]
            if item[0] == 'fill':
                object.fill = ColorRGBA().from_sexpr(item[1])
                object.fill.precision = 4
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'property':
                if item[1] == 'Sheet name' or item[1] == 'Sheetname': object.sheetName = Property().from_sexpr(item)
                if item[1] == 'Sheet file' or item[1] == 'Sheetfile': object.fileName = Property().from_sexpr(item)
            if item[0] == 'pin': object.pins.append(HierarchicalPin().from_sexpr(item))
            if item[0] == 'instances':
                for instance in item[1:]:
                    object.instances.append(HierarchicalSheetProjectInstance.from_sexpr(instance))
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        fa = ' (fields_autoplaced)' if self.fieldsAutoplaced else ''

        expression =  f'{indents}(sheet (at {self.position.X} {self.position.Y}) (size {self.width} {self.height}){fa}\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += f'{indents}  (fill {self.fill.to_sexpr()})\n'
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += self.sheetName.to_sexpr(indent+2)
        expression += self.fileName.to_sexpr(indent+2)
        for pin in self.pins:
            expression += pin.to_sexpr(indent+2)
        if len(self.instances) != 0:
            expression += f'{indents}  (instances\n'
            for instance in self.instances:
                expression += instance.to_sexpr(indent+4)
            expression += f'{indents}  )\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class HierarchicalSheetInstance():
    """The sheet_instance token defines the per sheet information for the entire schematic. This
       section will only exist in schematic files that are the root sheet of a project

    Documentation:
           https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_hierarchical_sheet_instance_section
    """

    instancePath: str = "/"
    """The ``instancePath`` attribute is the path to the sheet instance"""

    page: str = "1"
    """The ``page`` token defines the page number of the schematic represented by the sheet
       instance information. Page numbers can be any valid string."""

    @classmethod
    def from_sexpr(cls, exp: list) -> HierarchicalSheetInstance:
        """Convert the given S-Expresstion into a HierarchicalSheetInstance object

        Args:
            - exp (list): Part of parsed S-Expression ``(path ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not path

        Returns:
            - HierarchicalSheetInstance: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'path':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.instancePath = exp[1]
        for item in exp[2:]:
            if item[0] == 'page': object.page = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        return f'{indents}(path "{dequote(self.instancePath)}" (page "{dequote(self.page)}")){endline}'

@dataclass
class SymbolInstance():
    """The ``symbol_instance`` token defines the per symbol information for the entire schematic

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-schematic/#_symbol_instance_section
    """

    path: str = "/"
    """The ``path`` attribute is the path to the sheet instance"""

    reference: str = ""
    """The ``reference`` token attribute is a string that defines the reference designator for
       the symbol instance"""

    unit: int = 0
    """The unit token attribute is a integer ordinal that defines the symbol unit for the
       symbol instance. For symbols that do not define multiple units, this will always be 1."""

    value: str = ""
    """The value token attribute is a string that defines the value field for the symbol instance"""

    footprint: str = ""
    """The ``footprint`` token attribute is a string that defines the LIBRARY_IDENTIFIER for footprint associated with the symbol instance"""

    @classmethod
    def from_sexpr(cls, exp: list) -> SymbolInstance:
        """Convert the given S-Expresstion into a SymbolInstance object

        Args:
            - exp (list): Part of parsed S-Expression ``(path ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not path

        Returns:
            - SymbolInstance: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'path':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.path = exp[1]
        for item in exp[2:]:
            if item[0] == 'reference': object.reference = item[1]
            if item[0] == 'unit': object.unit = item[1]
            if item[0] == 'value': object.value = item[1]
            if item[0] == 'footprint': object.footprint = item[1]
        return object

    def to_sexpr(self, indent=4, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 4.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(path "{dequote(self.path)}"\n'
        expression += f'{indents}  (reference "{dequote(self.reference)}") (unit {self.unit}) (value "{dequote(self.value)}") (footprint "{dequote(self.footprint)}")\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Rectangle():
    """The ``rectangle`` token defines a graphical rectangle in a schematic.

    Available since KiCad v7

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_symbol_rectangle
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token attributes define the coordinates of the start point of the rectangle"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token attributes define the coordinates of the end point of the rectangle"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the rectangle outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how rectangle arc is filled"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Rectangle:
        """Convert the given S-Expresstion into a Rectangle object

        Args:
            - exp (list): Part of parsed S-Expression ``(rectangle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not rectangle

        Returns:
            - Rectangle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'rectangle':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            if item[0] == 'end': object.end = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(rectangle (start {self.start.X} {self.start.Y}) (end {self.end.X} {self.end.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Arc():
    """The ``Arc`` token defines a graphical arc in a schematic.

    Available since KiCad v7

    Documentation:
        - ???
    """

    start: Position = field(default_factory=lambda: Position())
    """The ``start`` token attributes define the coordinates of the start point of the arc"""

    mid: Position = field(default_factory=lambda: Position())
    """The ``end`` token attributes define the coordinates of the mid point of the arc"""

    end: Position = field(default_factory=lambda: Position())
    """The ``end`` token attributes define the coordinates of the end point of the arc"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the arc outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how the arc is filled"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Arc:
        """Convert the given S-Expresstion into a Arc object

        Args:
            - exp (list): Part of parsed S-Expression ``(arc ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not arc

        Returns:
            - Arc: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'arc':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if item[0] == 'start': object.start = Position().from_sexpr(item)
            if item[0] == 'mid': object.mid = Position().from_sexpr(item)
            if item[0] == 'end': object.end = Position().from_sexpr(item)
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(arc (start {self.start.X} {self.start.Y}) (mid {self.mid.X} {self.mid.Y}) (end {self.end.X} {self.end.Y})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class Circle():
    """The ``Circle`` token defines a graphical circle in a schematic.

    Available since KiCad v7

    Documentation:
        - ???
    """

    center: Position = field(default_factory=lambda: Position())
    """The ``center`` token attributes define the coordinates of the center point of the circle"""

    radius: float = 0.0
    """The ``radius`` token attributes define the radius of the circle"""

    stroke: Stroke = field(default_factory=lambda: Stroke())
    """The ``stroke`` defines how the circle outline is drawn"""

    fill: Fill = field(default_factory=lambda: Fill())
    """The ``fill`` token attributes define how the circle is filled"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier. Defaults to ``None.``"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Circle:
        """Convert the given S-Expresstion into a Circle object

        Args:
            - exp (list): Part of parsed S-Expression ``(circle ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not circle

        Returns:
            - Circle: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'circle':
            raise Exception("Expression does not have the correct type")

        object = cls()

        for item in exp:
            if item[0] == 'center': object.center = Position().from_sexpr(item)
            if item[0] == 'radius': object.radius = item[1]
            if item[0] == 'stroke': object.stroke = Stroke().from_sexpr(item)
            if item[0] == 'fill': object.fill = Fill().from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        expression =  f'{indents}(circle (center {self.center.X} {self.center.Y}) (radius {self.radius})\n'
        expression += self.stroke.to_sexpr(indent+2)
        expression += self.fill.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        expression += f'{indents}){endline}'
        return expression
    
@dataclass
class NetclassFlag():
    """The ``netclass_flag`` token defines a netclass flag in a schematic.

    Available since KiCad v7

    Documentation:
        - ???
    """

    text: str = ""
    """The ``text`` token defines the text the netclass flag"""

    length: float = 2.54
    """The ``length`` token defines the length of the netclass flag"""

    shape: str = "round"
    """The ``shape`` token defines the shape of the netclass flag. Valid values are ``round``,
    ``rectangle``, ``dot`` or``diamond``."""

    position: Position = field(default_factory=lambda: Position)
    """The ``position`` token defines the position and rotation of the netclass flag"""

    effects: Effects = field(default_factory=lambda: Effects)
    """The ``effects`` token defines how the text is drawn"""

    properties: List[Property] = field(default_factory=list)
    """The ``properties`` token defines a list of properties the netclass is assigned to"""

    uuid: Optional[str] = None
    """The optional ``uuid`` defines the universally unique identifier"""

    fieldsAutoplaced: bool = False
    """The ``fields_autoplaced`` is a flag that indicates that any PROPERTIES associated
    with the netclas flag have been place automatically"""

    @classmethod
    def from_sexpr(cls, exp: list) -> NetclassFlag:
        """Convert the given S-Expresstion into a Circle object

        Args:
            - exp (list): Part of parsed S-Expression ``(netclass_flag ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not netclass_flag

        Returns:
            - NetclassFlag: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'netclass_flag':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.text = exp[1]
        for item in exp[2:]:
            if item[0] == 'length': object.length = item[1]
            if item[0] == 'shape': object.shape = item[1]
            if item[0] == 'at': object.position = Position.from_sexpr(item)
            if item[0] == 'fields_autoplaced': object.fieldsAutoplaced = True
            if item[0] == 'effects': object.effects = Effects.from_sexpr(item)
            if item[0] == 'uuid': object.uuid = item[1]
            if item[0] == 'property': object.properties.append(Property.from_sexpr(item))
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        posA = f' {self.position.angle}' if self.position.angle is not None else ''
        fa = f' (fields_autoplaced)' if self.fieldsAutoplaced else ''

        expression =  f'{indents}(netclass_flag "{dequote(self.text)}" (length {self.length}) (shape {self.shape}) (at {self.position.X} {self.position.Y}{posA}){fa}\n'
        expression += self.effects.to_sexpr(indent+2)
        if self.uuid is not None:
            expression += f'{indents}  (uuid {self.uuid})\n'
        for property in self.properties:
            expression += property.to_sexpr(indent+2)
        expression += f'{indents}){endline}'
        return expression


"""Unittests of footprint related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.footprint import Footprint

FOOTPRINT_BASE = path.join(TEST_BASE, 'footprint')

class Tests_Footprint(unittest.TestCase):
    """Test cases for Footprints"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_allFootprintItems(self):
        """Tests the parsing of all available footprint items"""
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_allFootprintItems')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_footprintPadNewLines(self):
        """Tests the ability of the parser to create the correct new-line breaks when using
        exotic footprint pad combinations"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_footprintPadNewLines')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_footprintEmptyAttributes(self):
        """Tests the correct parsing of a footprint with empty ``Attributes`` field

        </p><b>Note:</b> Some earlier versions of KiCad seemed to include the ``(attr ..)`` token in footprints even
        when no attributes are set (or when using standard attributes). This test includes an empty
        attribute token in the footprint and expects it to be gone after parsing. </p>
        """
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_footprintEmptyAttributes')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_createNewFootprintTypeSMD(self):
        """Tests the ``create_new()`` function to create a new footprint with the type ``smd``"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_createNewFootprintTypeSMD')

        # Create footprint with correct type
        footprint = Footprint().create_new(
            type = 'smd',
            library_id = 'empty-footprint-smd',
            value = 'empty-footprint-smd'
        )

        # Set timestamps to be the same as in the expected test output
        footprint.tedit = '6328915F'

        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_createNewFootprintTypeTHT(self):
        """Tests the ``create_new()`` function to create a new footprint with the type ``through_hole``"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_createNewFootprintTypeTHT')

        # Create footprint with correct type
        footprint = Footprint().create_new(
            type = 'through_hole',
            library_id = 'empty-footprint-through_hole',
            value = 'empty-footprint-through_hole'
        )

        # Set timestamps to be the same as in the expected test output
        footprint.tedit = '63289145'

        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_createNewFootprintTypeOther(self):
        """Tests the ``create_new()`` function to create a new footprint with the type ``other``"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'test_createNewFootprintTypeOther')

        # Create footprint with correct type
        footprint = Footprint().create_new(
            type = 'other',
            library_id = 'empty-footprint-other',
            value = 'empty-footprint-other'
        )
        # Set timestamps to be the same as in the expected test output
        footprint.tedit = '6328916A'

        self.assertTrue(to_file_and_compare(footprint, self.testData))

class Tests_Footprint_Since_V7(unittest.TestCase):
    """Test cases for Footprints since KiCad 7"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_textBoxAllVariants(self):
        """Tests all variants of the ``text_box`` token for text boxes in footprints"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'since_v7', 'test_textBoxAllVariants')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))

    def test_imageWithLayerToken(self):
        """Tests the new ``layer`` token for images in footprints"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'since_v7', 'test_imageWithLayerToken')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))
        
    def test_textsWithRenderCaches(self):
        """Tests text elements with the ``render_cache`` token"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(FOOTPRINT_BASE, 'since_v7', 'test_textsWithRenderCaches')
        footprint = Footprint().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(footprint, self.testData))


"""Unittests of schematic related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path
from kiutils.items.schitems import HierarchicalSheetInstance

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.schematic import Schematic
from kiutils.items.common import Property

SCHEMATIC_BASE = path.join(TEST_BASE, 'schematic')

class Tests_Schematic(unittest.TestCase):
    """Test cases for Schematics"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_addPropertyToSchematicSymbol(self):
        """Adds a new property to an already existing symbol in the schematic and verifies the
        correct initial values for the Property() class."""
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_addPropertyToSchematicSymbol')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        schematic.schematicSymbols[0].properties.append(
            Property(key='Property3', value='I was added from "outside" of KiCad', id=6)
        )
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_createEmptySchematic(self):
        """Tests that an empty schematic generates S-Expression as expected from KiCad

        Note: This test currently disregards an empty ``(symbol_instances)`` token as it seems that this
        only exists when empty KiCad schematics are created. This is what should be expected for
        empty schematics:

        <pre><code>
        (kicad_sch (version 20211123) (generator kicad-python-tools)
            (paper "A4")
            (lib_symbols)
            (symbol_instances)
        )
        </code</pre>"""
        schematic = Schematic.create_new()
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_createEmptySchematic')
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_schematicWithAllPrimitives(self):
        """Tests the parsing of a schematic with all primitives (lines, traces, busses, connections,
        images, etc)"""
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_schematicWithAllPrimitives')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_hierarchicalSchematicWithAllPrimitives(self):
        """Tests the parsing of a hierarchical schematic with all primitives (lines, traces, busses,
        connections, images, etc)"""
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_hierarchicalSchematicWithAllPrimitives')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_renameSymbolIdTokenInSchematic(self):
        """Tests if renaming (setting and unsetting) schematic symbols as well as normal symbols 
        using their ID token works as expected. Checks that the ``Value`` property does not change."""
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_renameSymbolIdTokenInSchematic')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        schematic.libSymbols[0].libId = "RenamedSwitch:SW_Coded_New"        # Setting library nickname
        schematic.libSymbols[1].libId = "Unset_Lib_Id"                      # Unsetting library nickname
        schematic.schematicSymbols[0].libId = "SwitchRenamed:SW_Coded_2"    # Setting library nickname
        schematic.schematicSymbols[1].libId = "Unset_Lib_Id"                # Unsetting library nickname
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_setSymbolLibNameToken(self):
        """Tests if setting and unsetting the lib_name token generates the correct S-Expression"""
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_setSymbolLibNameToken')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        schematic.schematicSymbols[0].libName = f"{schematic.schematicSymbols[0].entryName}_1"
        schematic.schematicSymbols[1].libName = None
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_parseStrokeTokens(self):
        """Tests the correct parsing of the Stroke token (with and without the color token)
         
        See:
            https://github.com/mvnmgrx/kiutils/pull/57
        """
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'test_parseStrokeTokens')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

class Tests_Schematic_Since_V7(unittest.TestCase):
    """Schematic related test cases since KiCad 7"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_textBoxAllVariants(self):
        """Tests all variants of the ``text_box`` token for text boxes in schematics"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_textBoxAllVariants')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_rectangleAllVariants(self):
        """Tests all variants of the ``rectangle`` token for rectangles in schematics"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_rectangleAllVariants')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_circleAllVariants(self):
        """Tests all variants of the ``circle`` token for circles in schematics"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_circleAllVariants')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_arcAllVariants(self):
        """Tests all variants of the ``arc`` token for arcs in schematics"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_arcAllVariants')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_schematicWithAllPrimitives(self):
        """Tests the parsing of a schematic with all primitives (lines, traces, busses, connections,
        images, etc) for KiCad 7"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_schematicWithAllPrimitives')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_netclassFlags(self):
        """Tests the parsing netclass flags for KiCad 7"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_netclassFlags')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_symbolPinOptionalTokens(self):
        """Tests the parsing of the optional name and number effects on symbol pins since KiCad v7.
        Came up in PR #73."""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_symbolPinOptionalTokens')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))

    def test_strokeOptionalTokens(self):
        """Tests the parsing of the optional tokens on strokes since KiCad v7.
        Came up in PR #73."""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(SCHEMATIC_BASE, 'since_v7', 'test_strokeOptionalTokens')
        schematic = Schematic().from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(schematic, self.testData))



"""Classes to manage KiCad footprints

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    02.02.2022 - created

Documentation taken from:
    https://dev-docs.kicad.org/en/file-formats/sexpr-footprint/
"""

from __future__ import annotations

import calendar
import datetime
import re
from dataclasses import dataclass, field
from typing import Optional, List, Dict
from os import path

from kiutils.items.zones import Zone
from kiutils.items.common import Image, Position, Coordinate, Net, Group, Font
from kiutils.items.fpitems import *
from kiutils.items.gritems import *
from kiutils.utils import sexpr
from kiutils.utils.strings import dequote, remove_prefix
from kiutils.misc.config import KIUTILS_CREATE_NEW_VERSION_STR

@dataclass
class Attributes():
    """The ``attr`` token defines the list of attributes of a footprint.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_attributes
    """

    type: Optional[str] = None
    """The optional ``type`` token defines the type of footprint. Valid footprint types are ``smd`` and
    ``through_hole``. May be none when no attributes are set."""

    boardOnly: bool = False
    """The optional ``boardOnly`` token indicates that the footprint is only defined in the board
    and has no reference to any schematic symbol"""

    excludeFromPosFiles: bool = False
    """The optional ``excludeFromPosFiles`` token indicates that the footprint position information
    should not be included when creating position files"""

    excludeFromBom: bool = False
    """The optional ``excludeFromBom`` token indicates that the footprint should be excluded when
    creating bill of materials (BOM) files"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Attributes:
        """Convert the given S-Expresstion into a Attributes object

        Args:
            - exp (list): Part of parsed S-Expression ``(attr ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not attr

        Returns:
            - Attributes: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'attr':
            raise Exception("Expression does not have the correct type")

        object = cls()
        if len(exp) > 1:
            # Attributes token may be set with no other items (empty attributes)
            # Test case for this: test_fp_empty_attr.kicad_mod
            if exp[1] == 'through_hole' or exp[1] == 'smd':
                object.type = exp[1]

        for item in exp:
            if item == 'board_only': object.boardOnly = True
            if item == 'exclude_from_pos_files': object.excludeFromPosFiles = True
            if item == 'exclude_from_bom': object.excludeFromBom = True
        return object

    def to_sexpr(self, indent=0, newline=False) -> str:
        """Generate the S-Expression representing this object. Will return an empty string, if the
        following attributes are selected:
        - ``type``: None
        - ``boardOnly``: False
        - ``excludeFromBom``: False
        - ``excludeFromPosFiles``: False

        KiCad won't add the ``(attr ..)`` token to a footprint when this combination is selected.

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        if (self.type == None
            and self.boardOnly == False
            and self.excludeFromBom == False
            and self.excludeFromPosFiles == False):
            return ''

        indents = ' '*indent
        endline = '\n' if newline else ''
        type = f' {self.type}' if self.type is not None else ''

        expression = f'{indents}(attr{type}'
        if self.boardOnly is not None:
            if self.boardOnly:
                expression += ' board_only'
        if self.excludeFromPosFiles is not None:
            if self.excludeFromPosFiles:
                expression += ' exclude_from_pos_files'
        if self.excludeFromBom is not None:
            if self.excludeFromBom:
                expression += ' exclude_from_bom'
        expression += f'){endline}'
        return expression

@dataclass
class Model():
    """The ``model`` token defines the 3D model associated with a footprint.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_3d_model
    """

    path: str = ""
    """The ``path`` attribute is the path and file name of the 3D model"""

    pos: Coordinate = field(default_factory=lambda: Coordinate(0.0, 0.0, 0.0))
    """The ``pos`` token specifies the 3D position coordinates of the model relative to the footprint"""

    scale: Coordinate = field(default_factory=lambda: Coordinate(1.0, 1.0, 1.0))
    """The ``scale`` token specifies the model scale factor for each 3D axis"""

    rotate: Coordinate = field(default_factory=lambda: Coordinate(0.0, 0.0, 0.0))
    """The ``rotate`` token specifies the model rotation for each 3D axis relative to the footprint"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Model:
        """Convert the given S-Expresstion into a Model object

        Args:
            - exp (list): Part of parsed S-Expression ``(model ...)``

        Raises:
            - Exception: When given parameter's type is not a list or the list is not 5 long
            - Exception: When the first item of the list is not model

        Returns:
            - Model: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list) or len(exp) != 5:
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'model':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.path = exp[1]
        object.pos = Coordinate.from_sexpr(exp[2][1])
        object.scale = Coordinate.from_sexpr(exp[3][1])
        object.rotate = Coordinate.from_sexpr(exp[4][1])
        return object

    def to_sexpr(self, indent=2, newline=True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        expression =  f'{indents}(model "{dequote(self.path)}"\n'
        expression += f'{indents}  (offset {self.pos.to_sexpr()})\n'
        expression += f'{indents}  (scale {self.scale.to_sexpr()})\n'
        expression += f'{indents}  (rotate {self.rotate.to_sexpr()})\n'
        expression += f'{indents}){endline}'
        return expression

@dataclass
class DrillDefinition():
    """The ``drill`` token defines the drill attributes for a footprint pad.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_pad_drill_definition
    """

    oval: bool = False
    """The ``oval`` token defines if the drill is oval instead of round"""

    diameter: float = 0.0
    """The ``diameter`` attribute defines the drill diameter"""

    width: Optional[float] = None
    """The optional ``width`` attribute defines the width of the slot for oval drills"""

    offset: Optional[Position] = None
    """The optional ``offset`` token defines the drill offset coordinates from the center of the pad"""

    @classmethod
    def from_sexpr(cls, exp: list) -> DrillDefinition:
        """Convert the given S-Expresstion into a DrillDefinition object

        Args:
            - exp (list): Part of parsed S-Expression ``(drill ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not drill

        Returns:
            - DrillDefinition: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'drill':
            raise Exception("Expression does not have the correct type")

        object = cls()
        # Depending on the ``oval`` token, the fields may be shifted ..
        if exp[1] == 'oval':
            object.oval = True
            object.diameter = exp[2]
            object.width = exp[3]
        else:
            object.diameter = exp[1]
            if len(exp) > 2:
                object.width = exp[2]

        # The ``offset`` token may not be given
        for item in exp:
            if type(item) != type([]): continue
            if item[0] == 'offset': object.offset = Position().from_sexpr(item)
        return object

    def to_sexpr(self, indent: int = 0, newline: bool = False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        oval = f' oval' if self.oval else ''
        width = f' {self.width}' if self.oval and self.width is not None else ''
        offset = f' (offset {self.offset.X} {self.offset.Y})' if self.offset is not None else ''

        return f'{indents}(drill{oval} {self.diameter}{width}{offset}){endline}'

@dataclass
class PadOptions():
    """The ``options`` token attributes define the settings used for custom pads. This token is
    only used when a custom pad is defined.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_custom_pad_options
    """
    clearance: str = "outline"
    """The ``clearance`` token defines the type of clearance used for a custom pad. Valid clearance
    types are ``outline`` and ``convexhull``."""

    anchor: str = "rect"
    """The ``anchor`` token defines the anchor pad shape of a custom pad. Valid anchor pad shapes
    are rect and circle."""

    @classmethod
    def from_sexpr(cls, exp: list) -> PadOptions:
        """Convert the given S-Expresstion into a PadOptions object

        Args:
            - exp (list): Part of parsed S-Expression ``(options ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not options

        Returns:
            - PadOptions: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'options':
            raise Exception("Expression does not have the correct type")

        object = cls()
        for item in exp:
            if item[0] == 'clearance': object.clearance = item[1]
            if item[0] == 'anchor': object.anchor = item[1]
        return object

    def to_sexpr(self, indent: int = 0, newline: bool = False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to False.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        return f'{indents}(options (clearance {self.clearance}) (anchor {self.anchor})){endline}'

@dataclass
class Pad():
    """The ``pad`` token defines a pad in a footprint definition.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint_pad
    """

    number: str = "x"
    """The ``number`` attribute is the pad number"""

    type: str = "smd"
    """The pad ``type`` can be defined as ``thru_hole``, ``smd``, ``connect``, or ``np_thru_hole``"""

    shape: str = "rect"
    """The pad ``shape`` can be defined as ``circle``, ``rect``, ``oval``, ``trapezoid``, ``roundrect``, or
    ``custom``"""

    position: Position = field(default_factory=lambda: Position())
    """The ``position`` defines the X and Y coordinates and optional orientation angle of the pad"""

    locked: bool = False
    """The optional ``locked`` token defines if the footprint pad can be edited"""

    size: Position = field(default_factory=lambda: Position())         # Size uses Position class for simplicity for now
    """The ``size`` token defines the width and height of the pad"""

    drill: Optional[DrillDefinition] = None
    """The optional pad ``drill`` token defines the pad drill requirements"""

    # TODO: Test case for one-layer pad??
    layers: List[str] = field(default_factory=list)
    """The ``layers`` token defines the layer or layers the pad reside on"""

    property: Optional[str] = None
    """The optional ``property`` token defines any special properties for the pad. Valid properties
    are ``pad_prop_bga``, ``pad_prop_fiducial_glob``, ``pad_prop_fiducial_loc``, ``pad_prop_testpoint``,
    ``pad_prop_heatsink``, ``pad_prop_heatsink``, and ``pad_prop_castellated``"""

    removeUnusedLayers: bool = False
    """The optional ``removeUnusedLayers`` token specifies that the copper should be removed from
    any layers the pad is not connected to"""

    keepEndLayers: bool = False
    """The optional ``keepEndLayers`` token specifies that the top and bottom layers should be
    retained when removing the copper from unused layers"""

    roundrectRatio: Optional[float] = None
    """The optional ``roundrectRatio`` token defines the scaling factor of the pad to corner radius
    for rounded rectangular and chamfered corner rectangular pads. The scaling factor is a
    number between 0 and 1."""

    chamferRatio: Optional[float] = None   # Adds a newline before
    """The optional ``chamferRatio`` token defines the scaling factor of the pad to chamfer size.
    The scaling factor is a number between 0 and 1."""

    chamfer: List[str] = field(default_factory=list)
    """The optional ``chamfer`` token defines a list of one or more rectangular pad corners that
    get chamfered. Valid chamfer corner attributes are ``top_left``, ``top_right``, ``bottom_left``,
    and ``bottom_right``."""

    net: Optional[Net] = None
    """The optional ``net`` token defines the integer number and name string of the net connection
    for the pad."""

    tstamp: Optional[str] = None           # Used since KiCad 6
    """The optional ``tstamp`` token defines the unique identifier of the pad object"""

    pinFunction: Optional[str] = None
    """The optional ``pinFunction`` token attribute defines the associated schematic symbol pin name"""

    pinType: Optional[str] = None
    """The optional ``pinType`` token attribute defines the associated schematic pin electrical type"""

    dieLength: Optional[float] = None      # Adds a newline before
    """The optional ``dieLength`` token attribute defines the die length between the component pad
    and physical chip inside the component package"""

    solderMaskMargin: Optional[float] = None
    """The optional ``solderMaskMargin`` token attribute defines the distance between the pad and
    the solder mask for the pad. If not set, the footprint solder_mask_margin is used."""

    solderPasteMargin: Optional[float] = None
    """The optional ``solderPasteMargin`` token attribute defines the distance the solder paste
    should be changed for the pad"""

    solderPasteMarginRatio: Optional[float] = None
    """The optional ``solderPasteMarginRatio`` token attribute defines the percentage to reduce the
    pad outline by to generate the solder paste size"""

    clearance: Optional[float] = None
    """The optional ``clearance`` token attribute defines the clearance from all copper to the pad.
    If not set, the footprint clearance is used."""

    zoneConnect: Optional[int] = None
    """The optional ``zoneConnect`` token attribute defines type of zone connect for the pad. If
    not defined, the footprint zone_connection setting is used. Valid connection types are
    integers values from 0 to 3 which defines:
    - 0: Pad is not connect to zone
    - 1: Pad is connected to zone using thermal relief
    - 2: Pad is connected to zone using solid fill
    - 3: Only through hold pad is connected to zone using thermal relief
    """

    thermalWidth: Optional[float] = None
    """The optional ``thermalWidth`` token attribute defines the thermal relief spoke width used for
    zone connection for the pad. This only affects a pad connected to a zone with a thermal
    relief. If not set, the footprint thermal_width setting is used."""

    thermalGap: Optional[float] = None
    """The optional ``thermalGap`` token attribute defines the distance from the pad to the zone of
    the thermal relief connection for the pad. This only affects a pad connected to a zone
    with a thermal relief. If not set, the footprint thermal_gap setting is used."""

    customPadOptions: Optional[PadOptions] = None
    """The optional ``customPadOptions`` token defines the options when a custom pad is defined"""

    # Documentation seems wrong about primitives here. It seems like its just a list
    # of graphical objects, but the docu suggests, besides the list, two other params
    # for the primitive token: width and fill
    # These two however are note generated under the primitive token from the KiCad
    # generator. These two params may be found in gr_poly or gr_XX only.
    # So for now, the custom pad primitives are only a list of graphical objects
    customPadPrimitives: List = field(default_factory=list)
    """The optional ``customPadPrimitives`` defines the drawing objects and options used to define
    a custom pad"""

    @classmethod
    def from_sexpr(cls, exp: list) -> Pad:
        """Convert the given S-Expresstion into a Pad object

        Args:
            - exp (list): Part of parsed S-Expression ``(pad ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not pad

        Returns:
            - Pad: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'pad':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.number = exp[1]
        object.type = exp[2]
        object.shape = exp[3]

        for item in exp[3:]:
            if type(item) != type([]):
                if item == 'locked': object.locked = True

            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'size': object.size = Position().from_sexpr(item)
            if item[0] == 'drill': object.drill = DrillDefinition().from_sexpr(item)
            if item[0] == 'layers':
                for layer in item[1:]:
                    object.layers.append(layer)
            if item[0] == 'property': object.property = item[1]
            if item[0] == 'remove_unused_layers': object.removeUnusedLayers = True
            if item[0] == 'keep_end_layers': object.keepEndLayers = True
            if item[0] == 'roundrect_rratio': object.roundrectRatio = item[1]
            if item[0] == 'chamfer_ratio': object.chamferRatio = item[1]
            if item[0] == 'chamfer':
                for chamfer in item[1:]:
                    object.chamfer.append(chamfer)
            if item[0] == 'net': object.net = Net().from_sexpr(item)
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'pinfunction': object.pinFunction = item[1]
            if item[0] == 'pintype': object.pinType = item[1]
            if item[0] == 'die_length': object.dieLength = item[1]
            if item[0] == 'solder_mask_margin': object.solderMaskMargin = item[1]
            if item[0] == 'solder_paste_margin': object.solderPasteMargin = item[1]
            if item[0] == 'solder_paste_margin_ratio': object.solderPasteMarginRatio = item[1]
            if item[0] == 'clearance': object.clearance = item[1]
            if item[0] == 'zone_connect': object.zoneConnect = item[1]
            if item[0] == 'thermal_width': object.thermalWidth = item[1]
            if item[0] == 'thermal_gap': object.thermalGap = item[1]
            if item[0] == 'options': object.customPadOptions = PadOptions().from_sexpr(item)
            if item[0] == 'primitives':
                for primitive in item[1:]:
                    if primitive[0] == 'gr_text': object.customPadPrimitives.append(GrText().from_sexpr(primitive))
                    if primitive[0] == 'gr_text_box': object.customPadPrimitives.append(GrTextBox().from_sexpr(primitive))
                    if primitive[0] == 'gr_line': object.customPadPrimitives.append(GrLine().from_sexpr(primitive))
                    if primitive[0] == 'gr_rect': object.customPadPrimitives.append(GrRect().from_sexpr(primitive))
                    if primitive[0] == 'gr_circle': object.customPadPrimitives.append(GrCircle().from_sexpr(primitive))
                    if primitive[0] == 'gr_arc': object.customPadPrimitives.append(GrArc().from_sexpr(primitive))
                    if primitive[0] == 'gr_poly': object.customPadPrimitives.append(GrPoly().from_sexpr(primitive))
                    if primitive[0] == 'gr_curve': object.customPadPrimitives.append(GrCurve().from_sexpr(primitive))

                    # XXX: Are dimentions even implemented here?
                    if primitive[0] == 'dimension': raise NotImplementedError("Dimensions are not yet handled! Please report this bug along with the file being parsed.")
        return object

    def to_sexpr(self, indent: int = 2, newline: bool = True) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 2.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''
        champferFound, marginFound, schematicSymbolAssociated = False, False, False
        c, cr, smm, spm, spmr, cl, zc, tw, tg = '', '', '', '', '', '', '', '', ''

        layers = ' (layers'
        for layer in self.layers:
            # For some reason KiCad does not escape a layer with double-quotes if it has a
            # wildcard (*) or an ampersant (&) in it
            if "*." in layer or "&" in layer:
                layers += f' {layer}'
            else:
                layers += f' "{dequote(layer)}"'

        layers += ')'

        locked = ' locked' if self.locked else ''
        drill = f' {self.drill.to_sexpr()}' if self.drill is not None else ''
        ppty = f' (property {self.property})' if self.property is not None else ''
        rul = ' (remove_unused_layers)' if self.removeUnusedLayers else ''
        kel = ' (keep_end_layers)' if self.keepEndLayers else ''
        rrr = f' (roundrect_rratio {self.roundrectRatio})' if self.roundrectRatio is not None else ''

        net = f' {self.net.to_sexpr()}' if self.net is not None else ''
        pf = f' (pinfunction "{dequote(self.pinFunction)}")' if self.pinFunction is not None else ''
        pt = f' (pintype "{dequote(self.pinType)}")' if self.pinType is not None else ''

        # Check if a schematic symbol is associated with this footprint. This is usually set, if the
        # footprint is used in a board file.
        if net != '' or pf != '' or pt != '':
            schematicSymbolAssociated = True

        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''

        if len(self.chamfer) > 0:
            champferFound = True
            c = ' (chamfer'
            for chamfer in self.chamfer:
                c += f' {chamfer}'
            c += ')'
        if self.chamferRatio is not None:
            champferFound = True
            cr = f' (chamfer_ratio {self.chamferRatio})'

        if self.position.angle is not None:
            position = f'(at {self.position.X} {self.position.Y} {self.position.angle})'
        else:
            position = f'(at {self.position.X} {self.position.Y})'

        if self.solderMaskMargin is not None:
            marginFound = True
            smm = f' (solder_mask_margin {self.solderMaskMargin})'

        if self.solderPasteMargin is not None:
            marginFound = True
            spm = f' (solder_paste_margin {self.solderPasteMargin})'

        if self.solderPasteMarginRatio is not None:
            marginFound = True
            spmr = f' (solder_paste_margin_ratio {self.solderPasteMarginRatio})'

        if self.clearance is not None:
            marginFound = True
            cl = f' (clearance {self.clearance})'

        if self.zoneConnect is not None:
            marginFound = True
            zc = f' (zone_connect {self.zoneConnect})'

        if self.thermalWidth is not None:
            marginFound = True
            tw = f' (thermal_width {self.thermalWidth})'

        if self.thermalGap is not None:
            marginFound = True
            tg = f' (thermal_gap {self.thermalGap})'

        expression =  f'{indents}(pad "{dequote(str(self.number))}" {self.type} {self.shape}{locked} {position} (size {self.size.X} {self.size.Y}){drill}{ppty}{layers}{rul}{kel}{rrr}'
        if champferFound:
            # Only one whitespace here as all temporary strings have at least one leading whitespace
            expression += f'\n{indents} {cr}{c}'

        if self.dieLength is not None:
            expression += f'\n{indents}  (die_length {self.dieLength})'

        if marginFound or schematicSymbolAssociated:
            # Only one whitespace here as all temporary strings have at least one leading whitespace
            expression += f'\n{indents} {net}{pf}{pt}{smm}{spm}{spmr}{cl}{zc}{tw}{tg}'

        if self.customPadOptions is not None:
            expression += f'\n{indents}  {self.customPadOptions.to_sexpr()}'

        if self.customPadPrimitives is not None:
            if len(self.customPadPrimitives) > 0:
                expression += f'\n{indents}  (primitives'
                for primitive in self.customPadPrimitives:
                    expression += f'\n{primitive.to_sexpr(newline=False,indent=indent+4)}'
                expression += f'\n{indents}  )'

        expression += f'{tstamp}){endline}'
        return expression

@dataclass
class Footprint():
    """The ``footprint`` token defines a footprint.

    Documentation:
        https://dev-docs.kicad.org/en/file-formats/sexpr-intro/index.html#_footprint
    """

    @property
    def libId(self) -> str:
        """The ``lib_id`` token defines the link to footprint library of the footprint.
        This only applies to footprints defined in the board file format, in a regular footprint 
        file this id defines the footprint's name. In ``kiutils``, the token is a combination of
        both the ``libraryNickname`` and ``entryName`` token. Setting the ``lib_id`` token will
        update those tokens accordingly.

        Returns:
            - Symbol id in the following format: ``<libraryNickname>:<entryName>`` or ``<entryName>``,
              if ``libraryNickname`` token is not set.
        """
        if self.libraryNickname:
            return f'{self.libraryNickname}:{self.entryName}'
        else:
            return f'{self.entryName}'

    @libId.setter
    def libId(self, symbol_id: str):
        """Sets the ``lib_id`` token and parses its contents into the ``libraryNickname`` and
        ``entryName`` token.

        Args:
            - symbol_id (str): The symbol id in the following format: ``<libraryNickname>:<entryName>``
              or only ``<entryName>``
        """
        parse_symbol_id = re.match(r"^(.+?):(.+?)$", symbol_id)
        if parse_symbol_id:
            self.libraryNickname = parse_symbol_id.group(1)
            self.entryName = parse_symbol_id.group(2)
        else:
            self.libraryNickname = None
            self.entryName = symbol_id

    libraryNickname: Optional[str] = None
    """The optional ``libraryNickname`` token defines which symbol library this symbol belongs to
    and is a part of the ``id`` token"""
    
    entryName: str = None
    """The ``entryName`` token defines the actual name of the symbol and is a part of the ``id`` 
    token"""

    version: Optional[str] = None
    """The ``version`` token attribute defines the symbol library version using the YYYYMMDD date format"""

    generator: Optional[str] = None
    """The ``generator`` token attribute defines the program used to write the file"""

    locked: bool = False
    """The optional ``locked`` token defines a flag to indicate the footprint cannot be edited"""

    placed: bool = False
    """The optional ``placed`` token defines a flag to indicate that the footprint has not been placed"""

    layer: str = "F.Cu"
    """The ``layer`` token defines the canonical layer the footprint is placed"""

    tedit: str = remove_prefix(hex(calendar.timegm(datetime.datetime.now().utctimetuple())), '0x')
    """The ``tedit`` token defines a the last time the footprint was edited"""

    tstamp: Optional[str] = None
    """The ``tstamp`` token defines the unique identifier for the footprint. This only applies
    to footprints defined in the board file format."""

    position: Optional[Position] = None
    """The ``position`` token defines the X and Y coordinates and rotational angle of the
    footprint. This only applies to footprints defined in the board file format."""

    description: Optional[str] = None
    """The optional ``description`` token defines a string containing the description of the footprint"""

    tags: Optional[str] = None
    """The optional ``tags`` token defines a string of search tags for the footprint"""

    properties: Dict = field(default_factory=dict)
    """The ``properties`` token defines dictionary of properties as key / value pairs where key being
    the name of the property and value being the description of the property"""

    path: Optional[str] = None
    """The ``path`` token defines the hierarchical path of the schematic symbol linked to the footprint.
    This only applies to footprints defined in the board file format."""

    autoplaceCost90: Optional[int] = None
    """The optional ``autoplaceCost90`` token defines the vertical cost of when using the automatic
    footprint placement tool. Valid values are integers 1 through 10. This only applies to footprints
    defined in the board file format."""

    autoplaceCost180: Optional[int] = None
    """The optional ``autoplaceCost180`` token defines the horizontal cost of when using the automatic
    footprint placement tool. Valid values are integers 1 through 10. This only applies to footprints
    defined in the board file format."""

    solderMaskMargin: Optional[float] = None
    """The optional ``solderMaskMargin`` token defines the solder mask distance from all pads in the
    footprint. If not set, the board solder_mask_margin setting is used."""

    solderPasteMargin: Optional[float] = None
    """The optional ``solderPasteMargin`` token defines the solder paste distance from all pads in
    the footprint. If not set, the board solder_paste_margin setting is used."""

    solderPasteRatio: Optional[float] = None
    """The optional ``solderPasteRatio`` token defines the percentage of the pad size used to define
    the solder paste for all pads in the footprint. If not set, the board solder_paste_ratio setting
    is used."""

    clearance: Optional[float] = None
    """The optional ``clearance`` token defines the clearance to all board copper objects for all pads
    in the footprint. If not set, the board clearance setting is used."""

    zoneConnect: Optional[int] = None
    """The optional ``zoneConnect`` token defines how all pads are connected to filled zone. If not
    defined, then the zone connect_pads setting is used. Valid connection types are integers values
    from 0 to 3 which defines:
      - 0: Pads are not connect to zone
      - 1: Pads are connected to zone using thermal reliefs
      - 2: Pads are connected to zone using solid fill
      - 3: Only through hold pads are connected to zone using thermal reliefs
    """

    thermalWidth: Optional[float] = None
    """The optional ``thermalWidth`` token defined the thermal relief spoke width used for zone connections
    for all pads in the footprint. This only affects pads connected to zones with thermal reliefs. If
    not set, the zone thermal_width setting is used."""

    thermalGap: Optional[float] = None
    """The optional ``thermalGap`` is the distance from the pad to the zone of thermal relief connections
    for all pads in the footprint. If not set, the zone thermal_gap setting is used. If not set, the
    zone thermal_gap setting is used."""

    attributes: Attributes = field(default_factory=lambda: Attributes())
    """The optional ``attributes`` section defines the attributes of the footprint"""

    # TODO: Type hinting for this list
    graphicItems: List = field(default_factory=list)
    """The ``graphic`` objects section is a list of one or more graphical objects in the footprint. 
    Possible items are defined in ``kiutils.items.fpitems``. At minimum the reference designator 
    and value text objects are defined. All other graphical objects are optional.

    The ``Image`` token is supported since KiCad v7 and must be added into this list when used."""

    pads: List[Pad] = field(default_factory=list)
    """The optional ``pads`` section is a list of pads in the footprint"""

    zones: List[Zone] = field(default_factory=list)
    """The optional ``zones`` section is a list of keep out zones in the footprint"""

    groups: List[Group] = field(default_factory=list)
    """The optional ``groups`` section is a list of grouped objects in the footprint"""

    models: List[Model] = field(default_factory=list)
    """The ``3D model`` section defines the 3D model object associated with the footprint"""

    filePath: Optional[str] = None
    """The ``filePath`` token defines the path-like string to the library file. Automatically set when
    ``self.from_file()`` is used. Allows the use of ``self.to_file()`` without parameters."""

    @classmethod
    def from_sexpr(cls, exp: list) -> Footprint:
        """Convert the given S-Expresstion into a Footprint object

        Args:
            - exp (list): Part of parsed S-Expression ``(footprint ...)``

        Raises:
            - Exception: When given parameter's type is not a list
            - Exception: When the first item of the list is not footprint

        Returns:
            - Footprint: Object of the class initialized with the given S-Expression
        """
        if not isinstance(exp, list):
            raise Exception("Expression does not have the correct type")

        if exp[0] != 'module' and exp[0] != 'footprint':
            raise Exception("Expression does not have the correct type")

        object = cls()
        object.libId = exp[1]
        for item in exp[2:]:
            if not isinstance(item, list):
                if item == 'locked': object.locked = True
                if item == 'placed': object.placed = True
                continue

            if item[0] == 'version': object.version = item[1]
            if item[0] == 'generator': object.generator = item[1]
            if item[0] == 'layer': object.layer = item[1]
            if item[0] == 'tedit': object.tedit = item[1]
            if item[0] == 'tstamp': object.tstamp = item[1]
            if item[0] == 'descr': object.description = item[1]
            if item[0] == 'tags': object.tags = item[1]
            if item[0] == 'path': object.path = item[1]
            if item[0] == 'at': object.position = Position().from_sexpr(item)
            if item[0] == 'autoplace_cost90': object.autoplaceCost90 = item[1]
            if item[0] == 'autoplace_cost180': object.autoplaceCost180 = item[1]
            if item[0] == 'solder_mask_margin': object.solderMaskMargin = item[1]
            if item[0] == 'solder_paste_margin': object.solderPasteMargin = item[1]
            if item[0] == 'solder_paste_ratio': object.solderPasteRatio = item[1]
            if item[0] == 'clearance': object.clearance = item[1]
            if item[0] == 'zone_connect': object.zoneConnect = item[1]
            if item[0] == 'thermal_width': object.thermalWidth = item[1]
            if item[0] == 'thermal_gap': object.thermalGap = item[1]
            if item[0] == 'attr': object.attributes = Attributes.from_sexpr(item)
            if item[0] == 'model': object.models.append(Model.from_sexpr(item))
            if item[0] == 'fp_text': object.graphicItems.append(FpText.from_sexpr(item))
            if item[0] == 'fp_text_box': object.graphicItems.append(FpTextBox.from_sexpr(item))
            if item[0] == 'fp_line': object.graphicItems.append(FpLine.from_sexpr(item))
            if item[0] == 'fp_rect': object.graphicItems.append(FpRect.from_sexpr(item))
            if item[0] == 'fp_circle': object.graphicItems.append(FpCircle.from_sexpr(item))
            if item[0] == 'fp_arc': object.graphicItems.append(FpArc.from_sexpr(item))
            if item[0] == 'fp_poly': object.graphicItems.append(FpPoly.from_sexpr(item))
            if item[0] == 'fp_curve': object.graphicItems.append(FpCurve.from_sexpr(item))
            if item[0] == 'image':object.graphicItems.append(Image.from_sexpr(item))
            if item[0] == 'pad': object.pads.append(Pad.from_sexpr(item))
            if item[0] == 'zone': object.zones.append(Zone.from_sexpr(item))
            if item[0] == 'property': object.properties.update({ item[1]: item[2] })
            if item[0] == 'group': object.groups.append(Group.from_sexpr(item))
            if item[0] == 'dimension':
                raise NotImplementedError("Dimensions are not yet handled! Please report this bug along with the file being parsed.")

        return object

    @classmethod
    def from_file(cls, filepath: str, encoding: Optional[str] = None) -> Footprint:
        """Load a footprint directly from a KiCad footprint file (`.kicad_mod`) and sets the
        ``self.filePath`` attribute to the given file path.

        Args:
            - filepath (str): Path or path-like object that points to the file
            - encoding (str, optional): Encoding of the input file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If the given path is not a file

        Returns:
            - Footprint: Object of the Footprint class initialized with the given KiCad footprint
        """
        if not path.isfile(filepath):
            raise Exception("Given path is not a file!")

        with open(filepath, 'r', encoding=encoding) as infile:
            rawFootprint = infile.read()

            fpData = sexpr.parse_sexp(rawFootprint)
            return cls.from_sexpr(fpData)

    @classmethod
    def create_new(cls, library_id: str, value: str,
                        type: str = 'other', reference: str = 'REF**') -> Footprint:
        """Creates a new empty footprint with its attributes set as KiCad would create it

        Args:
            - library_link (str): Denotes the name of the library as well as the footprint. Like `Connector:Conn01x02`)
            - value (str): The value text item (printed on the fabrication layer as ``value`` attribute)
            - type (str): Type of footprint (``smd``, ``through_hole`` or ``other``). Defaults to 'other'.
            - reference (str): Reference of the footprint. Defaults to `REF**`.
        Raises:
            - Exception: When the given type is something other than listed above

        Returns:
            - Footprint: Empty footprint
        """
        if type not in ['smd', 'through_hole', 'other']:
            raise Exception("Unsupported type was given")

        fp = cls(
            version = KIUTILS_CREATE_NEW_VERSION_STR,
            generator = 'kiutils'
        )
        fp.libId = library_id

        # Create text items that are created when adding a new footprint to a library
        fp.graphicItems.extend(
            [
                FpText(
                    type = 'reference', text = reference, layer = 'F.SilkS',
                    effects = Effects(font=Font(thickness=0.15)),
                    position = Position(X=0, Y=-0.5, unlocked=True)
                ),
                FpText(
                    type = 'value', text = value, layer ='F.Fab',
                    effects  = Effects(font=Font(thickness=0.15)),
                    position = Position(X=0, Y=1, unlocked=True)
                ),
                FpText(
                    type = 'user', text = '${REFERENCE}', layer = 'F.Fab',
                    effects = Effects(font=Font(thickness=0.15)),
                    position = Position(X=0, Y=2.5, unlocked=True)
                )
            ]
        )

        # The type ``other`` does not set the attributes type token
        if type != 'other':
            fp.attributes.type = type

        return fp

    def to_file(self, filepath = None, encoding: Optional[str] = None):
        """Save the object to a file in S-Expression format

        Args:
            - filepath (str, optional): Path-like string to the file. Defaults to None. If not set, 
                                        the attribute ``self.filePath`` will be used instead.
            - encoding (str, optional): Encoding of the output file. Defaults to None (platform 
                                        dependent encoding).

        Raises:
            - Exception: If no file path is given via the argument or via `self.filePath`
        """
        if filepath is None:
            if self.filePath is None:
                raise Exception("File path not set")
            filepath = self.filePath

        with open(filepath, 'w', encoding=encoding) as outfile:
            outfile.write(self.to_sexpr())

    def to_sexpr(self, indent=0, newline=True, layerInFirstLine=False) -> str:
        """Generate the S-Expression representing this object

        Args:
            - indent (int): Number of whitespaces used to indent the output. Defaults to 0.
            - newline (bool): Adds a newline to the end of the output. Defaults to True.
            - layerInFirstLine (bool): Prints the ``layer`` token in the first line. Defaults to False

        Returns:
            - str: S-Expression of this object
        """
        indents = ' '*indent
        endline = '\n' if newline else ''

        locked = ' locked' if self.locked else ''
        placed = ' placed' if self.placed else ''
        version = f' (version {self.version})' if self.version is not None else ''
        generator = f' (generator {self.generator})' if self.generator is not None else ''
        tstamp = f' (tstamp {self.tstamp})' if self.tstamp is not None else ''

        expression =  f'{indents}(footprint "{dequote(self.libId)}"{locked}{placed}{version}{generator}'
        if layerInFirstLine:
            expression += f' (layer "{dequote(self.layer)}")\n'
        else:
            expression += f'\n{indents}  (layer "{dequote(self.layer)}")\n'
        expression += f'{indents}  (tedit {self.tedit}){tstamp}\n'

        if self.position is not None:
            angle = f' {self.position.angle}' if self.position.angle is not None else ''
            expression += f'{indents}  (at {self.position.X} {self.position.Y}{angle})\n'
        if self.description is not None:
            expression += f'{indents}  (descr "{dequote(self.description)}")\n'
        if self.tags is not None:
            expression += f'{indents}  (tags "{dequote(self.tags)}")\n'
        for item in self.properties:
            expression += f'{indents}  (property "{dequote(item)}" "{dequote(self.properties[item])}")\n'
        if self.path is not None:
            expression += f'{indents}  (path "{dequote(self.path)}")\n'

        # Additional parameters used in board
        if self.autoplaceCost90 is not None:
            expression += f'{indents}  (autoplace_cost90 {self.autoplaceCost90})\n'
        if self.autoplaceCost180 is not None:
            expression += f'{indents}  (autoplace_cost180 {self.autoplaceCost180})\n'
        if self.solderMaskMargin is not None:
            expression += f'{indents}  (solder_mask_margin {self.solderMaskMargin})\n'
        if self.solderPasteMargin is not None:
            expression += f'{indents}  (solder_paste_margin {self.solderPasteMargin})\n'
        if self.solderPasteRatio is not None:
            expression += f'{indents}  (solder_paste_ratio {self.solderPasteRatio})\n'
        if self.clearance is not None:
            expression += f'{indents}  (clearance {self.clearance})\n'
        if self.zoneConnect is not None:
            expression += f'{indents}  (zone_connect {self.zoneConnect})\n'
        if self.thermalWidth is not None:
            expression += f'{indents}  (thermal_width {self.thermalWidth})\n'
        if self.thermalGap is not None:
            expression += f'{indents}  (thermal_gap {self.thermalGap})\n'

        if self.attributes is not None:
            # Note: If the attribute object has only standard values in it, it will return an
            #       empty string. Therefore, it should create its own newline and indentations only
            #       when needed.
            expression += self.attributes.to_sexpr(indent=indent+2, newline=True)

        for item in self.graphicItems:
            expression += item.to_sexpr(indent=indent+2)
        for item in self.pads:
            expression += item.to_sexpr(indent=indent+2)
        for item in self.zones:
            expression += item.to_sexpr(indent=indent+2)
        for item in self.models:
            expression += item.to_sexpr(indent=indent+2)
        for item in self.groups:
            expression += item.to_sexpr(indent=indent+2)

        expression += f'{indents}){endline}'
        return expression




"""Unittests of design-rules related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.dru import DesignRules

DESIGNRULE_BASE = path.join(TEST_BASE, 'designrules')

class Tests_DesignRules(unittest.TestCase):
    """Test cases for Design Rules"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_allDesignRuleItems(self):
        """Tests the parsing of all available design rule items"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(DESIGNRULE_BASE, 'test_allDesignRuleItems')
        dru = DesignRules.from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(dru, self.testData))

    def test_createNewDesignRules(self):
        """Tests the ``create_new()`` function to create a new set of design rules"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(DESIGNRULE_BASE, 'test_createNewDesignRules')
        dru = DesignRules.create_new()
        self.assertTrue(to_file_and_compare(dru, self.testData))

class Tests_DesignRules_Since_V7(unittest.TestCase):
    """Test cases for Design Rules since KiCad v7"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_severityToken(self):
        """Tests the parsing the new severity token"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(DESIGNRULE_BASE, 'since_v7', 'test_severityToken')
        dru = DesignRules.from_file(self.testData.pathToTestFile)
        self.assertTrue(to_file_and_compare(dru, self.testData))



"""Unittests of library table related classes

Authors:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0
"""

import unittest
from os import path

from tests.testfunctions import to_file_and_compare, prepare_test, cleanup_after_test, TEST_BASE
from kiutils.libraries import LibTable, Library

LIBTABLE_BASE = path.join(TEST_BASE, 'libtable')

class Tests_LibTable(unittest.TestCase):
    """Test cases for library tables"""

    def setUp(self) -> None:
        prepare_test(self)
        return super().setUp()

    def test_parseFpLibTable(self):
        """Tests the correct parsing of an example `fp-lib-table`"""
        self.testData.pathToTestFile = path.join(LIBTABLE_BASE, 'test_parseFpLibTable')
        libtable = LibTable().from_file(f'{self.testData.pathToTestFile}.expected')
        self.assertTrue(to_file_and_compare(libtable, self.testData))

    def test_parseSymLibTable(self):
        """Tests the correct parsing of an example `sym-lib-table`"""
        self.testData.pathToTestFile = path.join(LIBTABLE_BASE, 'test_parseSymLibTable')
        libtable = LibTable().from_file(f'{self.testData.pathToTestFile}.expected')
        self.assertTrue(to_file_and_compare(libtable, self.testData))

    def test_addLibraryObjectToLibTable(self):
        """Tests adding a library object to a library table"""
        self.testData.pathToTestFile = path.join(LIBTABLE_BASE, 'test_addLibraryObjectToLibTable')
        libtable = LibTable(type='fp_lib_table')
        libtable.libs.append(Library(
            name = 'object1',
            type = "KiCad",
            uri = '${KIPRJMOD}/my/library.pretty',
            options = 'Some options with "quoted strings"',
            description = 'Some description with "quoted strings"',
            active = False
        ))
        self.assertTrue(to_file_and_compare(libtable, self.testData))

    def test_addEmptyLibraryObjectToLibTable(self):
        """Tests adding an empty library object to a library table"""
        self.testData.pathToTestFile = path.join(LIBTABLE_BASE, 'test_addEmptyLibraryObjectToLibTable')
        libtable = LibTable(type='fp_lib_table')
        libtable.libs.append(Library())
        self.assertTrue(to_file_and_compare(libtable, self.testData))

    def test_createNewLibTable(self):
        """Tests the ``create_new()`` function to create a new library table"""
        self.testData.compareToTestFile = True
        self.testData.pathToTestFile = path.join(LIBTABLE_BASE, 'test_createNewLibTable')
        libtable = LibTable.create_new()
        self.assertTrue(to_file_and_compare(libtable, self.testData))

    # TODO: Tests with invalid type token